Option Strict On
Option Explicit On

Imports System.Drawing
Imports System.Drawing.Imaging
Imports System.Runtime.InteropServices

Public Class Form1
    ' Speichert die fertige Graustufenpalette
    Private Bmp8GrayPal As ColorPalette

    ''' <summary>
    ''' Konvertiert ein Bitmapobjekt in ein 8bppIndexedGrayScale Bitmapobjekt
    ''' </summary>
    ''' <param name="InBitmap">Bitmapobjekt</param>
    ''' <returns>8bppIndexedGrayScale Bitmapobjekt</returns>
    Private Function ConvertTo8bppIndexedGrayScale( _
        ByVal InBitmap As Bitmap) As Bitmap

        ' Fr die For/Next Schleifen
        Dim intX As New Integer
        Dim intY As New Integer

        ' Fr die Pixelpositionen im ByteArray
        Dim Bmp8Pos As New Integer
        Dim Bmp24Pos As New Integer

        ' Breite einer Bildzeile inkl. PadBytes berechnen
        Dim Bmp8Stride As Integer = (InBitmap.Width + 3) And Not 3
        Dim Bmp24Stride As Integer = ((InBitmap.Width * 3) + 3) And Not 3

        ' ByteArrays zur Aufnahme der gesamten Bitmapdaten dimensionieren
        Dim Bmp8Data As Byte() = _
            New Byte((InBitmap.Height * Bmp8Stride) - 1) {}
        Dim Bmp24Data As Byte() = _
            New Byte((InBitmap.Height * Bmp24Stride) - 1) {}

        ' Handle auf das gepinnte ByteArray Bmp24Data holen
        Dim hBmp24Data As GCHandle = _
            GCHandle.Alloc(Bmp24Data, GCHandleType.Pinned)

        ' InBitmap in ein 24bppRGB Bitmap konvertieren -> Bmp24
        ' Gleichzeitig stehen uns im gepinntem ByteArray Bmp24Data die 
        ' Bitmapdaten von Bmp24 zur Verfgung
        Using Bmp24 As New Bitmap(InBitmap.Width, InBitmap.Height, _
            Bmp24Stride, PixelFormat.Format24bppRgb, _
            hBmp24Data.AddrOfPinnedObject())

            ' Graphicsobjekt von Bmp24 erstellen -> Bmp24Gra
            Using Bmp24Gra As Graphics = Graphics.FromImage(Bmp24)

                ' InBitmap in das Graphicsobjekt zeichnen
                Bmp24Gra.DrawImageUnscaledAndClipped(InBitmap, _
                    New Rectangle(0, 0, Bmp24.Width, Bmp24.Height))

                ' Bmp24Gra lschen (dispose)
            End Using

            ' alle Pixel durchlaufen
            For intY = 0 To Bmp24.Height - 1
                For intX = 0 To Bmp24.Width - 1

                    ' Pixelpositionen in den ByteArrays berechnen
                    Bmp24Pos = (intY * Bmp24Stride) + (intX * 3)
                    Bmp8Pos = (intY * Bmp8Stride) + intX

                    ' RGB-Farbe zu Index in der 8bpp-Palette umrechnen
                    ' der berechnete Index verweist auf eine Farbe in der
                    ' Farbpalette -> Bmp8GrayPal
                    Bmp8Data(Bmp8Pos) = _
                        CByte(((77 * Bmp24Data(Bmp24Pos + 2)) + _
                            (150 * Bmp24Data(Bmp24Pos + 1)) + _
                            (28 * Bmp24Data(Bmp24Pos + 0))) \ 255)
                Next
            Next

            ' Bmp24 lschen (dispose)
        End Using

        ' Handle auf das gepinnte ByteArray Bmp24Data freigeben
        hBmp24Data.Free()

        ' Handle auf das gepinnte ByteArray Bmp8Data holen
        Dim hBmp8Data As GCHandle = _
            GCHandle.Alloc(Bmp8Data, GCHandleType.Pinned)

        ' Neues BitmapData-Objekt erstellen
        Dim NewBmp8BD As New BitmapData
        NewBmp8BD.Width = InBitmap.Width
        NewBmp8BD.Height = InBitmap.Height
        NewBmp8BD.Stride = Bmp8Stride
        NewBmp8BD.PixelFormat = PixelFormat.Format8bppIndexed
        ' Zeiger auf das gepinnte ByteArray
        NewBmp8BD.Scan0 = hBmp8Data.AddrOfPinnedObject()

        ' Neues 8bppIndexed Bitmap erstellen -> Bmp8
        Dim Bmp8 As New Bitmap(InBitmap.Width, InBitmap.Height, _
            PixelFormat.Format8bppIndexed)

        ' Bitmapdaten von Bmp8 sperren und die Daten von NewBmp8BD bertragen
        Dim Bmp8BD As BitmapData = _
            Bmp8.LockBits(New Rectangle(0, 0, Bmp8.Width, Bmp8.Height), _
            ImageLockMode.WriteOnly Or ImageLockMode.UserInputBuffer, _
            Bmp8.PixelFormat, NewBmp8BD)

        ' Sperrung der Bitmapdaten aufheben
        Bmp8.UnlockBits(Bmp8BD)

        ' Handle auf das gepinnte ByteArray Bmp8Data freigeben
        hBmp8Data.Free()

        ' Standard-Farbpalette durch die Graustufenpalette ersetzen
        Bmp8.Palette = Bmp8GrayPal

        Return Bmp8
    End Function

    Private Sub Button1_Click(ByVal sender As System.Object, _
        ByVal e As System.EventArgs) Handles Button1.Click

        ' ist ein Bild in der PictureBox vorhanden
        If Not PictureBox2.Image Is Nothing Then
            ' Bild lschen  
            PictureBox2.Image.Dispose()
        End If

        PictureBox2.Image = _
            ConvertTo8bppIndexedGrayScale(CType(PictureBox1.Image, Bitmap))
    End Sub

    Private Sub Form1_Load(ByVal sender As Object, _
        ByVal e As System.EventArgs) Handles Me.Load

        ' Fr die For/Next Scleife
        Dim intItem As New Integer

        ' 1x1 8bppIndexed-Bitmap temporr zum auslesen der Farbpalette erstellen
        Using tmpBmp8 As New Bitmap(1, 1, PixelFormat.Format8bppIndexed)

            ' aktuelle Farbpalette von tmpBmp8 auslesen
            ' diese enthlt zu diesem Zeitpunkt bereits
            ' eine Standard-Farbpalette die wir durch eine
            ' neue Graustufenfarbpalette ersetzen
            Bmp8GrayPal = tmpBmp8.Palette

            ' alle Paletteneintrge durchlaufen
            For intItem = 0 To Bmp8GrayPal.Entries.Length - 1

                ' aktuelle Farbe in der Palette durch eine neue
                ' Graustufenfarbe ersetzen
                Bmp8GrayPal.Entries(intItem) = _
                    Color.FromArgb(255, intItem, intItem, intItem)
            Next

            ' tmpBmp8 lschen
        End Using

        ' Die erzeugte Graustufenpalette kann mehrfach verwendet werden und muss
        ' so nicht jedesmal beim Konvertieren neu erstellt werden, was dem 
        ' Ganzen nochmal einen kleinen Geschwindigkeitsschub verleiht.

        ' Bild aus der Ressource laden
        PictureBox1.Image = My.Resources.City005
    End Sub
End Class
